from collections import namedtuple
from contextlib import contextmanager
from types import TracebackType
from typing import Any, Generator, List, IO, Optional, Tuple, Type, Union
import io
import logging
import os
import threading
import sys

# NOTE: This is the canonical location for commonly-used vendored modules,
# which is the only spot that performs this try/except to allow repackaged
# Invoke to function (e.g. distro packages which unvendor the vendored bits and
# thus must import our 'vendored' stuff from the overall environment.)
# All other uses of Lexicon, etc should do 'from .util import lexicon' etc.
# Saves us from having to update the same logic in a dozen places.
# TODO: would this make more sense to put _into_ invoke.vendor? That way, the
# import lines which now read 'from .util import <third party stuff>' would be
# more obvious. Requires packagers to leave invoke/vendor/__init__.py alone tho
try:
    from .vendor.lexicon import Lexicon  # noqa
    from .vendor import yaml  # noqa
except ImportError:
    from lexicon import Lexicon  # type: ignore[no-redef] # noqa
    import yaml  # type: ignore[no-redef] # noqa


LOG_FORMAT = "%(name)s.%(module)s.%(funcName)s: %(message)s"


def enable_logging() -> None:
    logging.basicConfig(level=logging.DEBUG, format=LOG_FORMAT)


# Allow from-the-start debugging (vs toggled during load of tasks module) via
# shell env var.
if os.environ.get("INVOKE_DEBUG"):
    enable_logging()

# Add top level logger functions to global namespace. Meh.
log = logging.getLogger("invoke")
debug = log.debug


def task_name_sort_key(name: str) -> Tuple[List[str], str]:
    """
    Return key tuple for use sorting dotted task names, via e.g. `sorted`.

    .. versionadded:: 1.0
    """
    parts = name.split(".")
    return (
        # First group/sort by non-leaf path components. This keeps everything
        # grouped in its hierarchy, and incidentally puts top-level tasks
        # (whose non-leaf path set is the empty list) first, where we want them
        parts[:-1],
        # Then we sort lexicographically by the actual task name
        parts[-1],
    )


# TODO: Make part of public API sometime
@contextmanager
def cd(where: str) -> Generator[None, None, None]:
    cwd = os.getcwd()
    os.chdir(where)
    try:
        yield
    finally:
        os.chdir(cwd)


def has_fileno(stream: IO) -> bool:
    """
    Cleanly determine whether ``stream`` has a useful ``.fileno()``.

    .. note::
        This function helps determine if a given file-like object can be used
        with various terminal-oriented modules and functions such as `select`,
        `termios`, and `tty`. For most of those, a fileno is all that is
        required; they'll function even if ``stream.isatty()`` is ``False``.

    :param stream: A file-like object.

    :returns:
        ``True`` if ``stream.fileno()`` returns an integer, ``False`` otherwise
        (this includes when ``stream`` lacks a ``fileno`` method).

    .. versionadded:: 1.0
    """
    try:
        return isinstance(stream.fileno(), int)
    except (AttributeError, io.UnsupportedOperation):
        return False


def isatty(stream: IO) -> Union[bool, Any]:
    """
    Cleanly determine whether ``stream`` is a TTY.

    Specifically, first try calling ``stream.isatty()``, and if that fails
    (e.g. due to lacking the method entirely) fallback to `os.isatty`.

    .. note::
        Most of the time, we don't actually care about true TTY-ness, but
        merely whether the stream seems to have a fileno (per `has_fileno`).
        However, in some cases (notably the use of `pty.fork` to present a
        local pseudoterminal) we need to tell if a given stream has a valid
        fileno but *isn't* tied to an actual terminal. Thus, this function.

    :param stream: A file-like object.

    :returns:
        A boolean depending on the result of calling ``.isatty()`` and/or
        `os.isatty`.

    .. versionadded:: 1.0
    """
    # If there *is* an .isatty, ask it.
    if hasattr(stream, "isatty") and callable(stream.isatty):
        return stream.isatty()
    # If there wasn't, see if it has a fileno, and if so, ask os.isatty
    elif has_fileno(stream):
        return os.isatty(stream.fileno())
    # If we got here, none of the above worked, so it's reasonable to assume
    # the darn thing isn't a real TTY.
    return False


def helpline(obj: object) -> Optional[str]:
    """
    Yield an object's first docstring line, or None if there was no docstring.

    .. versionadded:: 1.0
    """
    docstring = obj.__doc__
    if (
        not docstring
        or not docstring.strip()
        or docstring == type(obj).__doc__
    ):
        return None
    return docstring.lstrip().splitlines()[0]


class ExceptionHandlingThread(threading.Thread):
    """
    Thread handler making it easier for parent to handle thread exceptions.

    Based in part on Fabric 1's ThreadHandler. See also Fabric GH issue #204.

    When used directly, can be used in place of a regular ``threading.Thread``.
    If subclassed, the subclass must do one of:

    - supply ``target`` to ``__init__``
    - define ``_run()`` instead of ``run()``

    This is because this thread's entire point is to wrap behavior around the
    thread's execution; subclasses could not redefine ``run()`` without
    breaking that functionality.

    .. versionadded:: 1.0
    """

    def __init__(self, **kwargs: Any) -> None:
        """
        Create a new exception-handling thread instance.

        Takes all regular `threading.Thread` keyword arguments, via
        ``**kwargs`` for easier display of thread identity when raising
        captured exceptions.
        """
        super().__init__(**kwargs)
        # No record of why, but Fabric used daemon threads ever since the
        # switch from select.select, so let's keep doing that.
        self.daemon = True
        # Track exceptions raised in run()
        self.kwargs = kwargs
        # TODO: legacy cruft that needs to be removed
        self.exc_info: Optional[
            Union[
                Tuple[Type[BaseException], BaseException, TracebackType],
                Tuple[None, None, None],
            ]
        ] = None

    def run(self) -> None:
        try:
            # Allow subclasses implemented using the "override run()'s body"
            # approach to work, by using _run() instead of run(). If that
            # doesn't appear to be the case, then assume we're being used
            # directly and just use super() ourselves.
            # XXX https://github.com/python/mypy/issues/1424
            if hasattr(self, "_run") and callable(self._run):  # type: ignore
                # TODO: this could be:
                # - io worker with no 'result' (always local)
                # - tunnel worker, also with no 'result' (also always local)
                # - threaded concurrent run(), sudo(), put(), etc, with a
                # result (not necessarily local; might want to be a subproc or
                # whatever eventually)
                # TODO: so how best to conditionally add a "capture result
                # value of some kind"?
                # - update so all use cases use subclassing, add functionality
                # alongside self.exception() that is for the result of _run()
                # - split out class that does not care about result of _run()
                # and let it continue acting like a normal thread (meh)
                # - assume the run/sudo/etc case will use a queue inside its
                # worker body, orthogonal to how exception handling works
                self._run()  # type: ignore
            else:
                super().run()
        except BaseException:
            # Store for actual reraising later
            self.exc_info = sys.exc_info()
            # And log now, in case we never get to later (e.g. if executing
            # program is hung waiting for us to do something)
            msg = "Encountered exception {!r} in thread for {!r}"
            # Name is either target function's dunder-name, or just "_run" if
            # we were run subclass-wise.
            name = "_run"
            if "target" in self.kwargs:
                name = self.kwargs["target"].__name__
            debug(msg.format(self.exc_info[1], name))  # noqa

    def exception(self) -> Optional["ExceptionWrapper"]:
        """
        If an exception occurred, return an `.ExceptionWrapper` around it.

        :returns:
            An `.ExceptionWrapper` managing the result of `sys.exc_info`, if an
            exception was raised during thread execution. If no exception
            occurred, returns ``None`` instead.

        .. versionadded:: 1.0
        """
        if self.exc_info is None:
            return None
        return ExceptionWrapper(self.kwargs, *self.exc_info)

    @property
    def is_dead(self) -> bool:
        """
        Returns ``True`` if not alive and has a stored exception.

        Used to detect threads that have excepted & shut down.

        .. versionadded:: 1.0
        """
        # NOTE: it seems highly unlikely that a thread could still be
        # is_alive() but also have encountered an exception. But hey. Why not
        # be thorough?
        return (not self.is_alive()) and self.exc_info is not None

    def __repr__(self) -> str:
        # TODO: beef this up more
        return str(self.kwargs["target"].__name__)


# NOTE: ExceptionWrapper defined here, not in exceptions.py, to avoid circular
# dependency issues (e.g. Failure subclasses need to use some bits from this
# module...)
#: A namedtuple wrapping a thread-borne exception & that thread's arguments.
#: Mostly used as an intermediate between `.ExceptionHandlingThread` (which
#: preserves initial exceptions) and `.ThreadException` (which holds 1..N such
#: exceptions, as typically multiple threads are involved.)
ExceptionWrapper = namedtuple(
    "ExceptionWrapper", "kwargs type value traceback"
)



from queue import Queue, Empty
from threading import Thread, Event

from invoke import Responder, FailingResponder, ResponseNotAccepted


# NOTE: StreamWatcher is basically just an interface/protocol; no behavior to
# test of its own. So this file tests Responder primarily, and some subclasses.


class Responder_:
    def keeps_track_of_seen_index_per_thread(self):
        # Instantiate a single object which will be used in >1 thread
        r = Responder(pattern="foo", response="bar fight")  # meh
        # Thread body func allowing us to mimic actual IO thread behavior, with
        # Queues used in place of actual pipes/files
        def body(responder, in_q, out_q, finished):
            while not finished.is_set():
                try:
                    # NOTE: use nowait() so our loop is hot & can shutdown ASAP
                    # if finished gets set.
                    stream = in_q.get_nowait()
                    for response in r.submit(stream):
                        out_q.put_nowait(response)
                except Empty:
                    pass

        # Create two threads from that body func, and queues/etc for each
        t1_in, t1_out, t1_finished = Queue(), Queue(), Event()
        t2_in, t2_out, t2_finished = Queue(), Queue(), Event()
        t1 = Thread(target=body, args=(r, t1_in, t1_out, t1_finished))
        t2 = Thread(target=body, args=(r, t2_in, t2_out, t2_finished))
        # Start the threads
        t1.start()
        t2.start()
        try:
            stream = "foo fighters"
            # First thread will basically always work
            t1_in.put(stream)
            assert t1_out.get() == "bar fight"
            # Second thread get() will block/timeout if threadlocals aren't in
            # use, because the 2nd thread's copy of the responder will not have
            # its own index & will thus already be 'past' the `foo` in the
            # stream.
            t2_in.put(stream)
            assert t2_out.get(timeout=1) == "bar fight"
        except Empty:
            assert (
                False
            ), "Unable to read from thread 2 - implies threadlocal indices are broken!"  # noqa
        # Close up.
        finally:
            t1_finished.set()
            t2_finished.set()
            t1.join()
            t2.join()

    def yields_response_when_regular_string_pattern_seen(self):
        r = Responder(pattern="empty", response="handed")
        assert list(r.submit("the house was empty")) == ["handed"]

    def yields_response_when_regex_seen(self):
        r = Responder(pattern=r"tech.*debt", response="pay it down")
        response = r.submit("technically, it's still debt")
        assert list(response) == ["pay it down"]

    def multiple_hits_within_stream_yield_multiple_responses(self):
        r = Responder(pattern="jump", response="how high?")
        assert list(r.submit("jump, wait, jump, wait")) == ["how high?"] * 2

    def patterns_span_multiple_lines(self):
        r = Responder(pattern=r"call.*problem", response="So sorry")
        output = """
You only call me
when you have a problem
You never call me
Just to say hi
"""
        assert list(r.submit(output)) == ["So sorry"]


class FailingResponder_:
    def behaves_like_regular_responder_by_default(self):
        r = FailingResponder(
            pattern="ju[^ ]{2}", response="how high?", sentinel="lolnope"
        )
        assert list(r.submit("jump, wait, jump, wait")) == ["how high?"] * 2

    def raises_failure_exception_when_sentinel_detected(self):
        r = FailingResponder(
            pattern="ju[^ ]{2}", response="how high?", sentinel="lolnope"
        )
        # Behaves normally initially
        assert list(r.submit("jump")) == ["how high?"]
        # But then!
        try:
            r.submit("lolnope")
        except ResponseNotAccepted as e:
            message = str(e)
            # Expect useful bits in exception text
            err = "Didn't see pattern in {!r}".format(message)
            assert "ju[^ ]{2}" in message, err
            err = "Didn't see failure sentinel in {!r}".format(message)
            assert "lolnope" in message, err
        else:
            assert False, "Did not raise ResponseNotAccepted!"



import os
from typing import TYPE_CHECKING, Optional

from invoke import Collection, task, Exit

from invocations import ci, checks
from invocations.docs import docs, www, sites, watch_docs
from invocations.pytest import coverage as coverage_, test as test_
from invocations.packaging import vendorize, release

if TYPE_CHECKING:
    from invoke import Context


@task
def test(
    c: "Context",
    verbose: bool = False,
    color: bool = True,
    capture: str = "no",
    module: Optional[str] = None,
    k: Optional[str] = None,
    x: bool = False,
    opts: str = "",
    pty: bool = True,
) -> None:
    """
    Run pytest. See `invocations.pytest.test` for details.

    This is a simple wrapper around the abovementioned task, which makes a
    couple minor defaults changes appropriate for this particular test suite,
    such as:

    - setting ``capture=no`` instead of ``capture=sys``, as we do a very large
      amount of subprocess IO testing that even the ``sys``  capture screws up
    - setting ``verbose=False`` because we have a large number of tests and
      skipping verbose output by default is a ~20% time savings.)
    """
    # TODO: update test suite to use c.config.run.in_stream = False globally.
    # somehow.
    test_(
        c,
        verbose=verbose,
        color=color,
        capture=capture,
        module=module,
        k=k,
        x=x,
        opts=opts,
        pty=pty,
    )


# TODO: replace with invocations' once the "call truly local tester" problem is
# solved (see other TODOs). For now this is just a copy/paste/modify.
@task(help=test.help)  # type: ignore
def integration(
    c: "Context", opts: Optional[str] = None, pty: bool = True
) -> None:
    """
    Run the integration test suite. May be slow!
    """
    # Abort if no default shell on this system - implies some unusual dev
    # environment. Certain entirely-standalone tests will fail w/o it, even if
    # tests honoring config overrides (like the unit-test suite) don't.
    shell = c.config.global_defaults()["run"]["shell"]
    if not c.run("which {}".format(shell), hide=True, warn=True):
        err = "No {} on this system - cannot run integration tests! Try a container?"  # noqa
        raise Exit(err.format(shell))
    opts = opts or ""
    opts += " integration/"
    test(c, opts=opts, pty=pty)


@task
def coverage(
    c: "Context", report: str = "term", opts: str = "", codecov: bool = False
) -> None:
    """
    Run pytest in coverage mode. See `invocations.pytest.coverage` for details.
    """
    # Use our own test() instead of theirs.
    # Also add integration test so this always hits both.
    # (Not regression, since that's "weird" / doesn't really hit any new
    # coverage points)
    coverage_(
        c,
        report=report,
        opts=opts,
        tester=test,
        additional_testers=[integration],
        codecov=codecov,
    )


@task
def regression(c: "Context", jobs: int = 8) -> None:
    """
    Run an expensive, hard-to-test-in-pytest run() regression checker.

    :param int jobs: Number of jobs to run, in total. Ideally num of CPUs.
    """
    os.chdir("integration/_support")
    cmd = "seq {} | parallel -n0 --halt=now,fail=1 inv -c regression check"
    c.run(cmd.format(jobs))


ns = Collection(
    test,
    coverage,
    integration,
    regression,
    vendorize,
    release,
    www,
    docs,
    sites,
    watch_docs,
    ci,
    checks.blacken,
    checks,
)
ns.configure(
    {
        "blacken": {
            # Skip vendor, build dirs when blackening.
            # TODO: this is making it seem like I really do want an explicit
            # arg/conf-opt in the blacken task for "excluded paths"...ha
            "find_opts": "-and -not \( -path './invoke/vendor*' -or -path './build*' \)"  # noqa
        },
        "packaging": {
            "wheel": True,
            "check_desc": True,
            "changelog_file": os.path.join(
                www.configuration()["sphinx"]["source"], "changelog.rst"
            ),
        },
    }
)



from contextlib import contextmanager
from functools import wraps
from resource import getrusage, RUSAGE_SELF
import sys
import time


from pytest import skip


def current_cpu_usage():
    rusage = getrusage(RUSAGE_SELF)
    return rusage.ru_utime + rusage.ru_stime


@contextmanager
def assert_cpu_usage(lt, verbose=False):
    """
    Execute wrapped block, asserting CPU utilization was less than ``lt``%.

    :param float lt: CPU use percentage above which failure will occur.
    :param bool verbose: Whether to print out the calculated percentage.
    """
    start_usage = current_cpu_usage()
    start_time = time.time()
    yield
    end_usage = current_cpu_usage()
    end_time = time.time()

    usage_diff = end_usage - start_usage
    time_diff = end_time - start_time

    if time_diff == 0:  # Apparently possible!
        time_diff = 0.000001

    percentage = (usage_diff / time_diff) * 100.0

    if verbose:
        print("Used {0:.2}% CPU over {1:.2}s".format(percentage, time_diff))

    assert percentage < lt


def only_utf8(f):
    """
    Decorator causing tests to skip if local shell pipes aren't UTF-8.
    """
    # TODO: use actual test selection labels or whatever nose has
    @wraps(f)
    def inner(*args, **kwargs):
        if getattr(sys.stdout, "encoding", None) == "UTF-8":
            return f(*args, **kwargs)
        # TODO: could remove this so they show green, but figure yellow is more
        # appropriate
        skip()

    return inner



from invoke import Collection

# Issue #934 (from #919) only seems to trigger on this style of 'from . import
# xxx' - a vanilla self-contained tasks/__init__.py is still fine!
from . import module

ns = Collection(module)



from pytest import skip, raises

from invoke.parser import Argument


class Argument_:
    class init:
        "__init__"

        def may_take_names_list(self):
            names = ("--foo", "-f")
            a = Argument(names=names)
            # herp a derp
            for name in names:
                assert name in a.names

        def may_take_name_arg(self):
            assert "-b" in Argument(name="-b").names

        def must_get_at_least_one_name(self):
            with raises(TypeError):
                Argument()

        def default_arg_is_name_not_names(self):
            assert "b" in Argument("b").names

        def can_declare_positional(self):
            assert Argument(name="foo", positional=True).positional is True

        def positional_is_False_by_default(self):
            assert Argument(name="foo").positional is False

        def can_set_attr_name_to_control_name_attr(self):
            a = Argument("foo", attr_name="bar")
            assert a.name == "bar"  # not 'foo'

    class repr:
        "__repr__"

        def shows_useful_info(self):
            arg = Argument(names=("name", "nick1", "nick2"))
            expected = "<Argument: {} ({})>".format("name", "nick1, nick2")
            assert repr(arg) == expected

        def does_not_show_nickname_parens_if_no_nicknames(self):
            assert repr(Argument("name")) == "<Argument: name>"

        def shows_positionalness(self):
            arg = Argument("name", positional=True)
            assert repr(arg) == "<Argument: name *>"

        def shows_optionalness(self):
            arg = Argument("name", optional=True)
            assert repr(arg) == "<Argument: name ?>"

        def positionalness_and_optionalness_stick_together(self):
            # TODO: but do these even make sense on the same argument? For now,
            # best to have a nonsensical test than a missing one...
            arg = Argument("name", optional=True, positional=True)
            assert repr(arg) == "<Argument: name *?>"

        def shows_kind_if_not_str(self):
            assert repr(Argument("age", kind=int)) == "<Argument: age [int]>"

        def all_the_things_together(self):
            arg = Argument(
                names=("meh", "m"), kind=int, optional=True, positional=True
            )
            assert repr(arg) == "<Argument: meh (m) [int] *?>"

    class kind_kwarg:
        "'kind' kwarg"

        def is_optional(self):
            Argument(name="a")
            Argument(name="b", kind=int)

        def defaults_to_str(self):
            assert Argument("a").kind == str

        def non_bool_implies_value_needed(self):
            assert Argument(name="a", kind=int).takes_value
            assert Argument(name="b", kind=str).takes_value
            assert Argument(name="c", kind=list).takes_value

        def bool_implies_no_value_needed(self):
            assert not Argument(name="a", kind=bool).takes_value

        def bool_implies_default_False_not_None(self):
            # Right now, parsing a bool flag not given results in None
            # TODO: may want more nuance here -- False when a --no-XXX flag is
            # given, True if --XXX, None if not seen?
            # Only makes sense if we add automatic --no-XXX stuff (think
            # ./configure)
            skip()

        def may_validate_on_set(self):
            with raises(ValueError):
                Argument("a", kind=int).value = "five"

        def list_implies_initial_value_of_empty_list(self):
            assert Argument("mylist", kind=list).value == []

    class names:
        def returns_tuple_of_all_names(self):
            assert Argument(names=("--foo", "-b")).names == ("--foo", "-b")
            assert Argument(name="--foo").names == ("--foo",)

        def is_normalized_to_a_tuple(self):
            assert isinstance(Argument(names=("a", "b")).names, tuple)

    class name:
        def returns_first_name(self):
            assert Argument(names=("a", "b")).name == "a"

    class nicknames:
        def returns_rest_of_names(self):
            assert Argument(names=("a", "b")).nicknames == ("b",)

    class takes_value:
        def True_by_default(self):
            assert Argument(name="a").takes_value

        def False_if_kind_is_bool(self):
            assert not Argument(name="-b", kind=bool).takes_value

    class value_set:
        "value="

        def available_as_dot_raw_value(self):
            "available as .raw_value"
            a = Argument("a")
            a.value = "foo"
            assert a.raw_value == "foo"

        def untransformed_appears_as_dot_value(self):
            "untransformed, appears as .value"
            a = Argument("a", kind=str)
            a.value = "foo"
            assert a.value == "foo"

        def transformed_appears_as_dot_value_with_original_as_raw_value(self):
            "transformed, modified value is .value, original is .raw_value"
            a = Argument("a", kind=int)
            a.value = "5"
            assert a.value == 5
            assert a.raw_value == "5"

        def list_kind_triggers_append_instead_of_overwrite(self):
            # TODO: when put this way it makes the API look pretty strange;
            # maybe a sign we should switch to explicit setter methods
            # (selected on kind, perhaps) instead of using an implicit setter
            a = Argument("mylist", kind=list)
            assert a.value == []
            a.value = "val1"
            assert a.value == ["val1"]
            a.value = "val2"
            assert a.value == ["val1", "val2"]

        def incrementable_True_triggers_increment_of_default(self):
            a = Argument("verbose", kind=int, default=0, incrementable=True)
            assert a.value == 0
            # NOTE: parser currently just goes "Argument.takes_value is false?
            # Gonna stuff True/False in there." So this looks pretty silly out
            # of context (as with list-types above.)
            a.value = True
            assert a.value == 1
            for _ in range(4):
                a.value = True
            assert a.value == 5

    class value:
        def returns_default_if_not_set(self):
            a = Argument("a", default=25)
            assert a.value == 25

    class raw_value:
        def is_None_when_no_value_was_actually_seen(self):
            a = Argument("a", kind=int)
            assert a.raw_value is None

    class got_value:
        def non_list_kind_tests_for_None_value(self):
            arg = Argument("a")
            assert not arg.got_value
            arg.value = "something"
            assert arg.got_value

        def list_kind_test_for_empty_list_value(self):
            arg = Argument("a", kind=list)
            assert not arg.got_value
            arg.value = "append-me"
            assert arg.got_value

    class set_value:
        def casts_by_default(self):
            a = Argument("a", kind=int)
            a.set_value("5")
            assert a.value == 5

        def allows_setting_value_without_casting(self):
            a = Argument("a", kind=int)
            a.set_value("5", cast=False)
            assert a.value == "5"



import os
import sys
from importlib.machinery import ModuleSpec
from importlib.util import module_from_spec, spec_from_file_location
from pathlib import Path
from types import ModuleType
from typing import Any, Optional, Tuple

from . import Config
from .exceptions import CollectionNotFound
from .util import debug


class Loader:
    """
    Abstract class defining how to find/import a session's base `.Collection`.

    .. versionadded:: 1.0
    """

    def __init__(self, config: Optional["Config"] = None) -> None:
        """
        Set up a new loader with some `.Config`.

        :param config:
            An explicit `.Config` to use; it is referenced for loading-related
            config options. Defaults to an anonymous ``Config()`` if none is
            given.
        """
        if config is None:
            config = Config()
        self.config = config

    def find(self, name: str) -> Optional[ModuleSpec]:
        """
        Implementation-specific finder method seeking collection ``name``.

        Must return a ModuleSpec valid for use by `importlib`, which is
        typically a name string followed by the contents of the 3-tuple
        returned by `importlib.module_from_spec` (``name``, ``loader``,
        ``origin``.)

        For a sample implementation, see `.FilesystemLoader`.

        .. versionadded:: 1.0
        """
        raise NotImplementedError

    def load(self, name: Optional[str] = None) -> Tuple[ModuleType, str]:
        """
        Load and return collection module identified by ``name``.

        This method requires a working implementation of `.find` in order to
        function.

        In addition to importing the named module, it will add the module's
        parent directory to the front of `sys.path` to provide normal Python
        import behavior (i.e. so the loaded module may load local-to-it modules
        or packages.)

        :returns:
            Two-tuple of ``(module, directory)`` where ``module`` is the
            collection-containing Python module object, and ``directory`` is
            the string path to the directory the module was found in.

        .. versionadded:: 1.0
        """
        if name is None:
            name = self.config.tasks.collection_name
        spec = self.find(name)
        if spec and spec.loader and spec.origin:
            # Typically either tasks.py or tasks/__init__.py
            source_file = Path(spec.origin)
            # Will be 'the dir tasks.py is in', or 'tasks/', in both cases this
            # is what wants to be in sys.path for "from . import sibling"
            enclosing_dir = source_file.parent
            # Will be "the directory above the spot that 'import tasks' found",
            # namely the parent of "your task tree", i.e. "where project level
            # config files are looked for". So, same as enclosing_dir for
            # tasks.py, but one more level up for tasks/__init__.py...
            module_parent = enclosing_dir
            if spec.parent:  # it's a package, so we have to go up again
                module_parent = module_parent.parent
            # Get the enclosing dir on the path
            enclosing_str = str(enclosing_dir)
            if enclosing_str not in sys.path:
                sys.path.insert(0, enclosing_str)
            # Actual import
            module = module_from_spec(spec)
            sys.modules[spec.name] = module  # so 'from . import xxx' works
            spec.loader.exec_module(module)
            # Return the module and the folder it was found in
            return module, str(module_parent)
        msg = "ImportError loading {!r}, raising ImportError"
        debug(msg.format(name))
        raise ImportError


class FilesystemLoader(Loader):
    """
    Loads Python files from the filesystem (e.g. ``tasks.py``.)

    Searches recursively towards filesystem root from a given start point.

    .. versionadded:: 1.0
    """

    # TODO: could introduce config obj here for transmission to Collection
    # TODO: otherwise Loader has to know about specific bits to transmit, such
    # as auto-dashes, and has to grow one of those for every bit Collection
    # ever needs to know
    def __init__(self, start: Optional[str] = None, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        if start is None:
            start = self.config.tasks.search_root
        self._start = start

    @property
    def start(self) -> str:
        # Lazily determine default CWD if configured value is falsey
        return self._start or os.getcwd()

    def find(self, name: str) -> Optional[ModuleSpec]:
        debug("FilesystemLoader find starting at {!r}".format(self.start))
        spec = None
        module = "{}.py".format(name)
        paths = self.start.split(os.sep)
        try:
            # walk the path upwards to check for dynamic import
            for x in reversed(range(len(paths) + 1)):
                path = os.sep.join(paths[0:x])
                if module in os.listdir(path):
                    spec = spec_from_file_location(
                        name, os.path.join(path, module)
                    )
                    break
                elif name in os.listdir(path) and os.path.exists(
                    os.path.join(path, name, "__init__.py")
                ):
                    basepath = os.path.join(path, name)
                    spec = spec_from_file_location(
                        name,
                        os.path.join(basepath, "__init__.py"),
                        submodule_search_locations=[basepath],
                    )
                    break
            if spec:
                debug("Found module: {!r}".format(spec))
                return spec
        except (FileNotFoundError, ModuleNotFoundError):
            msg = "ImportError loading {!r}, raising CollectionNotFound"
            debug(msg.format(name))
            raise CollectionNotFound(name=name, start=self.start)
        return None



if input("What's the password?") == "Rosebud":
    print("You're not Citizen Kane!")
    # This should sit around forever like e.g. a bad sudo prompt would, but the
    # responder ought to be looking for the above and aborting instead.
    input("Seriously, what's the password???")



import re

from unittest.mock import patch

import invoke
import invoke.collection
import invoke.exceptions
import invoke.tasks
import invoke.program


class Init:
    "__init__"

    def dunder_version_info(self):
        assert hasattr(invoke, "__version_info__")
        ver = invoke.__version_info__
        assert isinstance(ver, tuple)
        assert all(isinstance(x, int) for x in ver)

    def dunder_version(self):
        assert hasattr(invoke, "__version__")
        ver = invoke.__version__
        assert isinstance(ver, str)
        assert re.match(r"\d+\.\d+\.\d+", ver)

    def dunder_version_looks_generated_from_dunder_version_info(self):
        # Meh.
        ver_part = invoke.__version__.split(".")[0]
        ver_info_part = invoke.__version_info__[0]
        assert ver_part == str(ver_info_part)

    class exposes_bindings:
        def task_decorator(self):
            assert invoke.task is invoke.tasks.task

        def task_class(self):
            assert invoke.Task is invoke.tasks.Task

        def collection_class(self):
            assert invoke.Collection is invoke.collection.Collection

        def context_class(self):
            assert invoke.Context is invoke.context.Context

        def mock_context_class(self):
            assert invoke.MockContext is invoke.context.MockContext

        def config_class(self):
            assert invoke.Config is invoke.config.Config

        def pty_size_function(self):
            assert invoke.pty_size is invoke.terminals.pty_size

        def local_class(self):
            assert invoke.Local is invoke.runners.Local

        def runner_class(self):
            assert invoke.Runner is invoke.runners.Runner

        def promise_class(self):
            assert invoke.Promise is invoke.runners.Promise

        def failure_class(self):
            assert invoke.Failure is invoke.runners.Failure

        def exceptions(self):
            # Meh
            for obj in vars(invoke.exceptions).values():
                if isinstance(obj, type) and issubclass(obj, BaseException):
                    top_level = getattr(invoke, obj.__name__)
                    real = getattr(invoke.exceptions, obj.__name__)
                    assert top_level is real

        def runner_result(self):
            assert invoke.Result is invoke.runners.Result

        def watchers(self):
            assert invoke.StreamWatcher is invoke.watchers.StreamWatcher
            assert invoke.Responder is invoke.watchers.Responder
            assert invoke.FailingResponder is invoke.watchers.FailingResponder

        def program(self):
            assert invoke.Program is invoke.program.Program

        def filesystemloader(self):
            assert invoke.FilesystemLoader is invoke.loader.FilesystemLoader

        def argument(self):
            assert invoke.Argument is invoke.parser.Argument

        def parsercontext(self):
            assert invoke.ParserContext is invoke.parser.ParserContext

        def parser(self):
            assert invoke.Parser is invoke.parser.Parser

        def parseresult(self):
            assert invoke.ParseResult is invoke.parser.ParseResult

        def executor(self):
            assert invoke.Executor is invoke.executor.Executor

        def call(self):
            assert invoke.call is invoke.tasks.call

        def Call(self):
            # Starting to think we shouldn't bother with lowercase-c call...
            assert invoke.Call is invoke.tasks.Call

    class offers_singletons:
        @patch("invoke.Context")
        def run(self, Context):
            result = invoke.run("foo", bar="biz")
            ctx = Context.return_value
            ctx.run.assert_called_once_with("foo", bar="biz")
            assert result is ctx.run.return_value

        @patch("invoke.Context")
        def sudo(self, Context):
            result = invoke.sudo("foo", bar="biz")
            ctx = Context.return_value
            ctx.sudo.assert_called_once_with("foo", bar="biz")
            assert result is ctx.sudo.return_value



"Tasks for managing Sphinx docs."

from invoke import task, Collection


@task(name="all", default=True)
def all_(c):
    "Build all doc formats."
    pass


@task
def html(c):
    "Build HTML output only."
    pass


@task
def pdf(c):
    "Build PDF output only."
    pass



from unittest.mock import Mock
from pytest import raises, skip

from invoke import Context, Config, task, Task, Call, Collection
from invoke import FilesystemLoader as Loader

from _util import support


#
# NOTE: Most Task tests use @task as it's the primary interface and is a very
# thin wrapper around Task itself. This way we don't have to write 2x tests for
# both Task and @task. Meh :)
#


def _func(c):
    pass


class task_:
    "@task"

    def _load(self, name):
        mod, _ = self.loader.load(name)
        return Collection.from_module(mod)

    def setup_method(self):
        self.loader = Loader(start=support)
        self.vanilla = self._load("decorators")

    def allows_access_to_wrapped_object(self):
        def lolcats(c):
            pass

        assert task(lolcats).body == lolcats

    def allows_alias_specification(self):
        assert self.vanilla["foo"] == self.vanilla["bar"]

    def allows_multiple_aliases(self):
        assert self.vanilla["foo"] == self.vanilla["otherbar"]

    def allows_default_specification(self):
        assert self.vanilla[""] == self.vanilla["biz"]

    def has_autoprint_option(self):
        ap = self._load("autoprint")
        assert ap["nope"].autoprint is False
        assert ap["yup"].autoprint is True

    def raises_ValueError_on_multiple_defaults(self):
        with raises(ValueError):
            self._load("decorator_multi_default")

    def sets_arg_kind(self):
        skip()

    def sets_which_args_are_optional(self):
        assert self.vanilla["optional_values"].optional == ("myopt",)

    def allows_annotating_args_as_positional(self):
        assert self.vanilla["one_positional"].positional == ["pos"]
        assert self.vanilla["two_positionals"].positional == ["pos1", "pos2"]

    def allows_annotating_args_as_iterable(self):
        assert self.vanilla["iterable_values"].iterable == ["mylist"]

    def allows_annotating_args_as_incrementable(self):
        arg = self.vanilla["incrementable_values"]
        assert arg.incrementable == ["verbose"]

    def when_positional_arg_missing_all_non_default_args_are_positional(self):
        arg = self.vanilla["implicit_positionals"]
        assert arg.positional == ["pos1", "pos2"]

    def context_arguments_should_not_appear_in_implicit_positional_list(self):
        @task
        def mytask(c):
            pass

        assert len(mytask.positional) == 0

    def pre_tasks_stored_directly(self):
        @task
        def whatever(c):
            pass

        @task(pre=[whatever])
        def func(c):
            pass

        assert func.pre == [whatever]

    def allows_star_args_as_shortcut_for_pre(self):
        @task
        def pre1(c):
            pass

        @task
        def pre2(c):
            pass

        @task(pre1, pre2)
        def func(c):
            pass

        assert func.pre == (pre1, pre2)

    def disallows_ambiguity_between_star_args_and_pre_kwarg(self):
        @task
        def pre1(c):
            pass

        @task
        def pre2(c):
            pass

        with raises(TypeError):

            @task(pre1, pre=[pre2])
            def func(c):
                pass

    def sets_name(self):
        @task(name="foo")
        def bar(c):
            pass

        assert bar.name == "foo"

    def returns_Task_instances_by_default(self):
        @task
        def mytask(c):
            pass

        assert isinstance(mytask, Task)

    def klass_kwarg_allows_overriding_class_used(self):
        class MyTask(Task):
            pass

        @task(klass=MyTask)
        def mytask(c):
            pass

        assert isinstance(mytask, MyTask)

    def klass_kwarg_works_for_subclassers_without_kwargs(self):
        # I.e. the previous test doesn't catch this particular use case
        class MyTask(Task):
            pass

        def uses_MyTask(*args, **kwargs):
            kwargs.setdefault("klass", MyTask)
            return task(*args, **kwargs)

        @uses_MyTask
        def mytask(c):
            pass

        assert isinstance(mytask, MyTask)

    def unknown_kwargs_get_mad_at_Task_level(self):
        # NOTE: this was previously untested behavior. We actually just
        # modified HOW TypeError gets raised (Task constructor, implicitly, vs
        # explicitly in @task itself) but the end result is the same for anyone
        # not trying to be stringly typed based on exception message.
        with raises(TypeError):

            @task(whatever="man")
            def mytask(c):
                pass


class Task_:
    def has_useful_repr(self):
        i = repr(Task(_func))
        assert "_func" in i, "'func' not found in {!r}".format(i)
        e = repr(Task(_func, name="funky"))
        assert "funky" in e, "'funky' not found in {!r}".format(e)
        assert "_func" not in e, "'_func' unexpectedly seen in {!r}".format(e)

    def equality_testing(self):
        t1 = Task(_func, name="foo")
        t2 = Task(_func, name="foo")
        assert t1 == t2
        t3 = Task(_func, name="bar")
        assert t1 != t3

    def equality_testing_false_for_non_task_objects(self):
        t = Task(_func, name="foo")
        # No name attribute at all
        assert t != object()
        # Name attr, but not a Task
        class Named:
            name = "foo"

        assert t != Named()

    class function_like_behavior:
        # Things that help them eg show up in autodoc easier
        def inherits_module_from_body(self):
            mytask = Task(_func, name="funky")
            assert mytask.__module__ is _func.__module__

    class attributes:
        def has_default_flag(self):
            assert Task(_func).is_default is False

        def name_defaults_to_body_name(self):
            assert Task(_func).name == "_func"

        def can_override_name(self):
            assert Task(_func, name="foo").name == "foo"

    class callability:
        def setup_method(self):
            @task
            def foo(c):
                "My docstring"
                return 5

            self.task = foo

        def dunder_call_wraps_body_call(self):
            context = Context()
            assert self.task(context) == 5

        def errors_if_first_arg_not_Context(self):
            @task
            def mytask(c):
                pass

            with raises(TypeError):
                mytask(5)

        def errors_if_no_first_arg_at_all(self):
            with raises(TypeError):

                @task
                def mytask():
                    pass

        def tracks_times_called(self):
            context = Context()
            assert self.task.called is False
            self.task(context)
            assert self.task.called is True
            assert self.task.times_called == 1
            self.task(context)
            assert self.task.times_called == 2

        def wraps_body_docstring(self):
            assert self.task.__doc__ == "My docstring"

        def wraps_body_name(self):
            assert self.task.__name__ == "foo"

    class get_arguments:
        def setup_method(self):
            @task(positional=["arg_3", "arg1"], optional=["arg1"])
            def mytask(c, arg1, arg2=False, arg_3=5):
                pass

            self.task = mytask
            self.args = self.task.get_arguments()
            self.argdict = self._arglist_to_dict(self.args)

        def _arglist_to_dict(self, arglist):
            # This kinda duplicates Context.add_arg(x) for x in arglist :(
            ret = {}
            for arg in arglist:
                for name in arg.names:
                    ret[name] = arg
            return ret

        def _task_to_dict(self, task):
            return self._arglist_to_dict(task.get_arguments())

        def positional_args_come_first(self):
            assert self.args[0].name == "arg_3"
            assert self.args[1].name == "arg1"
            assert self.args[2].name == "arg2"

        def kinds_are_preserved(self):
            # Remember that the default 'kind' is a string.
            assert [x.kind for x in self.args] == [int, str, bool]

        def positional_flag_is_preserved(self):
            assert [x.positional for x in self.args] == [True, True, False]

        def optional_flag_is_preserved(self):
            assert [x.optional for x in self.args] == [False, True, False]

        def optional_prevents_bool_defaults_from_affecting_kind(self):
            # Re #416. See notes in the function under test for rationale.
            @task(optional=["myarg"])
            def mytask(c, myarg=False):
                pass

            arg = mytask.get_arguments()[0]
            assert arg.kind is str  # not bool!

        def optional_plus_nonbool_default_does_not_override_kind(self):
            @task(optional=["myarg"])
            def mytask(c, myarg=17):
                pass

            arg = mytask.get_arguments()[0]
            assert arg.kind is int  # not str!

        def turns_function_signature_into_Arguments(self):
            assert len(self.args), 3 == str(self.args)
            assert "arg2" in self.argdict

        def shortflags_created_by_default(self):
            assert "a" in self.argdict
            assert self.argdict["a"] is self.argdict["arg1"]

        def shortflags_dont_care_about_positionals(self):
            "Positionalness doesn't impact whether shortflags are made"
            for short, long_ in (("a", "arg1"), ("r", "arg2"), ("g", "arg-3")):
                assert self.argdict[short] is self.argdict[long_]

        def autocreated_short_flags_can_be_disabled(self):
            @task(auto_shortflags=False)
            def mytask(c, arg):
                pass

            args = self._task_to_dict(mytask)
            assert "a" not in args
            assert "arg" in args

        def autocreated_shortflags_dont_collide(self):
            "auto-created short flags don't collide"

            @task
            def mytask(c, arg1, arg2, barg):
                pass

            args = self._task_to_dict(mytask)
            assert "a" in args
            assert args["a"] is args["arg1"]
            assert "r" in args
            assert args["r"] is args["arg2"]
            assert "b" in args
            assert args["b"] is args["barg"]

        def early_auto_shortflags_shouldnt_lock_out_real_shortflags(self):
            # I.e. "task --foo -f" => --foo should NOT get to pick '-f' for its
            # shortflag or '-f' is totally fucked.
            @task
            def mytask(c, longarg, l):  # noqa
                pass

            args = self._task_to_dict(mytask)
            assert "longarg" in args
            assert "o" in args
            assert args["o"] is args["longarg"]
            assert "l" in args

        def context_arguments_are_not_returned(self):
            @task
            def mytask(c):
                pass

            assert len(mytask.get_arguments()) == 0

        def underscores_become_dashes(self):
            @task
            def mytask(c, longer_arg):
                pass

            arg = mytask.get_arguments()[0]
            assert arg.names == ("longer-arg", "l")
            assert arg.attr_name == "longer_arg"
            assert arg.name == "longer_arg"

        class help:
            def setup_method(self):
                @task(
                    help={
                        "simple": "key",
                        "with_underscores": "yup",
                        "with-dashes": "also yup",
                    }
                )
                def mytask(c, simple, with_underscores, with_dashes):
                    pass

                self.help = {
                    arg.name: arg.help for arg in mytask.get_arguments()
                }

            def base_case(self):
                assert self.help["simple"] == "key"

            def underscored_name_via_underscores(self):
                assert self.help["with_underscores"] == "yup"

            def underscored_name_via_dashes(self):
                assert self.help["with_dashes"] == "also yup"

            def raises_ValueError_on_keys_not_found_in_task_args(self):
                @task(help={"non-existing-param": "Help text"})
                def no_parameters(c):
                    pass

                err = r"field was set.*don't exist:.*'non-existing-param'*"
                with raises(ValueError, match=err):
                    no_parameters.get_arguments()

            def no_ValueError_on_unfound_keys_when_configured_otherwise(self):
                @task(
                    help={"non-existing-param": "Help text", "param": "ganoes"}
                )
                def malazan(c, param):
                    pass

                arg_dict = {
                    arg.name: arg.help
                    for arg in malazan.get_arguments(ignore_unknown_help=True)
                }
                assert "non_existing_param" not in arg_dict
                assert arg_dict["param"] == "ganoes"

            def arg_value_is_copied_to_avoid_state_bleed_when_shared(self):
                # TODO: the 'real' solve here is to make sharing common
                # arguments between tasks a first class citizen; there's
                # tickets for that.
                shared_help = {"shared": "help"}

                @task(help=shared_help)
                def first(c, shared):
                    pass

                @task(help=shared_help)
                def second(c, shared):
                    pass

                # Running get_arguments() is required to trigger bug or lack
                # thereof, but also lets us do a nice safety check after.
                helps = [x.get_arguments()[0].help for x in (first, second)]
                assert helps == ["help", "help"]


# Dummy task for Call tests
_ = object()


class Call_:
    def setup_method(self):
        self.task = Task(Mock(__name__="mytask"))

    class init:
        class task:
            def is_required(self):
                with raises(TypeError):
                    Call()

            def is_first_posarg(self):
                assert Call(_).task is _

        class called_as:
            def defaults_to_None(self):
                assert Call(_).called_as is None

            def may_be_given(self):
                assert Call(_, called_as="foo").called_as == "foo"

        class args:
            def defaults_to_empty_tuple(self):
                assert Call(_).args == tuple()

            def may_be_given(self):
                assert Call(_, args=(1, 2, 3)).args == (1, 2, 3)

        class kwargs:
            def defaults_to_empty_dict(self):
                assert Call(_).kwargs == dict()

            def may_be_given(self):
                assert Call(_, kwargs={"foo": "bar"}).kwargs == {"foo": "bar"}

    class stringrep:
        "__str__"

        def includes_task_name(self):
            call = Call(self.task)
            assert str(call) == "<Call 'mytask', args: (), kwargs: {}>"

        def works_for_subclasses(self):
            class MyCall(Call):
                pass

            call = MyCall(self.task)
            assert "<MyCall" in str(call)

        def includes_args_and_kwargs(self):
            call = Call(
                self.task,
                args=("posarg1", "posarg2"),
                # Single-key dict to avoid dict ordering issues
                kwargs={"kwarg1": "val1"},
            )
            expected = "<Call 'mytask', args: ('posarg1', 'posarg2'), kwargs: {'kwarg1': 'val1'}>"  # noqa
            assert str(call) == expected

        def includes_aka_if_explicit_name_given(self):
            call = Call(self.task, called_as="notmytask")
            expected = "<Call 'mytask' (called as: 'notmytask'), args: (), kwargs: {}>"  # noqa
            assert str(call) == expected

        def skips_aka_if_explicit_name_same_as_task_name(self):
            call = Call(self.task, called_as="mytask")
            assert str(call) == "<Call 'mytask', args: (), kwargs: {}>"

    class make_context:
        def requires_config_argument(self):
            with raises(TypeError):
                Call(_).make_context()

        def creates_a_new_Context_from_given_config(self):
            conf = Config(defaults={"foo": "bar"})
            c = Call(_).make_context(conf)
            assert isinstance(c, Context)
            assert c.foo == "bar"

    class clone:
        def returns_new_but_equivalent_object(self):
            orig = Call(self.task)
            clone = orig.clone()
            assert clone is not orig
            assert clone == orig

        def can_clone_into_a_subclass(self):
            orig = Call(self.task)

            class MyCall(Call):
                pass

            clone = orig.clone(into=MyCall)
            assert clone == orig
            assert isinstance(clone, MyCall)

        def can_be_given_extra_kwargs_to_clone_with(self):
            orig = Call(self.task)

            class MyCall(Call):
                def __init__(self, *args, **kwargs):
                    self.hooray = kwargs.pop("hooray")
                    super().__init__(*args, **kwargs)

            clone = orig.clone(into=MyCall, with_={"hooray": "woo"})
            assert clone.hooray == "woo"



"""
Barebones regression-catching script that looks for ephemeral run() failures.

Intended to be run from top level of project via ``inv regression``. In an
ideal world this would be truly part of the integration test suite, but:

- something about the outer invoke or pytest environment seems to prevent such
  issues from appearing reliably (see eg issue #660)
- it can take quite a while to run, even compared to other integration tests.
"""


import sys

from invoke import task


@task
def check(c):
    count = 0
    failures = []
    for _ in range(0, 1000):
        count += 1
        try:
            # 'ls' chosen as an arbitrary, fast-enough-for-looping but
            # does-some-real-work example (where eg 'sleep' is less useful)
            response = c.run("ls", hide=True)
            if not response.ok:
                failures.append(response)
        except Exception as e:
            failures.append(e)
    if failures:
        print("run() FAILED {}/{} times!".format(len(failures), count))
        sys.exit(1)
    else:
        print("No failures detected after {} runs, A-OK".format(count))



"How to deploy our code and configs."

from invoke import task


@task(default=True)
def everywhere(c):
    "Deploy to all targets."
    pass


@task(aliases=["db_servers"])
def db(c):
    "Deploy to our database servers."
    pass


@task
def web(c):
    "Update and bounce the webservers."
    pass
